# ==================================================================
#            ______ _     _ _        __  __           _ 
#           |  ____(_)   | | |      |  \/  |         | |
#           | |__   _  __| | | ___  | \  / | ___   __| |
#           |  __| | |/ _` | |/ _ \ | |\/| |/ _ \ / _` |
#           | |    | | (_| | |  __/ | |  | | (_) | (_| |
#           |_|    |_|\__,_|_|\___| |_|  |_|\___/ \__,_|
#                                                        fidle.utils
# ==================================================================
# A simple module to host usefull functions for Fidle practical work
# Jean-Luc Parouty 2022

from genericpath import isdir
import os, sys
import string
import random
import numpy as np
from pathlib import Path
from glob import glob
import yaml


from collections.abc import Iterable
import datetime
from IPython.display import display,Image,Markdown,HTML

import fidle.config as config



__version__   = config.VERSION


# -------------------------------------------------------------
# Display md
# -------------------------------------------------------------

def subtitle(t):
    '''
    Display a markdown subtitle
    args:
        text to print
    return:
        none
    '''
    display(Markdown(f'<br>**{t}**'))
    

def display_md(text):
    '''
    Display a markdown text
    args:
        text to print
    return:
        none
    '''
    display(Markdown(text))


def display_html(text):
    '''
    Display a HTML text
    args:
        html to print
    return:
        none
    '''
    display(HTML(text))


def display_img(img):
    '''
    Display an image
    args:
        image to print
    return:
        none
    '''
    display(Image(img))


def np_print(*args, precision=3, linewidth=120):
    '''
    Print numpy vectors with a given precision
    args:
        *args : list of numpy vectors
        precision : print precision (3)
        linewidth : line width (120)
    return:
        nothing
    '''
    with np.printoptions(precision=precision, linewidth=linewidth):
        for a in args:
            print(a)


# -------------------------------------------------------------
# Format data
# -------------------------------------------------------------

def hdelay(delay):
    '''
    Returns a duration in human readable format, like 01:14:28 543ms
    delay can be timedelta or seconds
    args:
        delay : in seconds or as a timedelta
    reurns:
        strings
    '''
    if type(delay) is not datetime.timedelta:
        delay=datetime.timedelta(seconds=delay)
    sec = delay.total_seconds()
    hh = sec // 3600
    mm = (sec // 60) - (hh * 60)
    ss = sec - hh*3600 - mm*60
    ms = (sec - int(sec))*1000
    return f'{hh:02.0f}:{mm:02.0f}:{ss:02.0f} {ms:03.0f}ms'


def hsize(num, suffix='o'):
    '''
    Returns a size in human readable version, like 4.3G
    args:
        num: size in bytes
    return:
        String
    '''
    for unit in ['','K','M','G','T','P','E','Z']:
        if abs(num) < 1024.0:
            return f'{num:3.1f} {unit}{suffix}'
        num /= 1024.0
    return f'{num:.1f} Y{suffix}'



# -------------------------------------------------------------
# Folder cooking
# -------------------------------------------------------------


def mkdir(path):
    '''
    Create a subdirectory
    Mode is 0750, do nothing if exist
    args:
        path : directory to create
    return:
        none
    '''
    os.makedirs(path, mode=0o750, exist_ok=True)
    

def get_directory_size(path):
    """
    Return the directory size, but only 1 level
    args:
        path : directory path
    return:
        size in Mo
    """
    size=0
    for f in os.listdir(path):
        if os.path.isfile(path+'/'+f):
            size+=os.path.getsize(path+'/'+f)
    return size/(1024*1024)


# -------------------------------------------------------------
# Datasets cooking
# -------------------------------------------------------------


def shuffle_np_dataset(*data):
    """
    Shuffle a list of dataset
    args:
        *data : datasets
    return:
        *datasets mixed
    """
    print('Datasets have been shuffled.')
    p = np.random.permutation(len(data[0]))
    out = [ d[p] for d in data ]
    return out[0] if len(out)==1 else out


def rescale_dataset(*data, scale=1):
    '''
    Rescale numpy array with 'scale' factor
    args:
        *data : arrays
        scale : scale factor
    return:
        arrays of rescaled data
    '''
    print('Datasets have been resized with a factor ', scale)
    out = [ d[:int(scale*len(d))] for d in data ]
    return out[0] if len(out)==1 else out


def pick_dataset(*data,n=5):
    '''Return random subsets of n elements'''
    # ii = np.random.choice(range(len(data[0])), n)
    ii=np.random.choice( len(data[0]), n, replace=False)
    out = [ d[ii] for d in data ]
    return out[0] if len(out)==1 else out


# -------------------------------------------------------------
# Misc
# -------------------------------------------------------------


def update_progress(what,i,imax, redraw=False, verbosity=2, total_max=None):
    """
    Display a text progress bar, as :
    My progress bar : ############# 34%

    Args:
        what  : Progress bar name
        i     : Current progress
        imax  : Max value for i
        verbosity : progress bar verbosity (0: no bar, 1: one line, 2: full progress bar)
        
    Returns:
        nothing
    """
    if verbosity==0:   return
    if verbosity==1 and i<imax: return
    if total_max is None : total_max = imax 
    bar_length = min(40,imax)
    if (i%int(imax/bar_length))!=0 and i<imax and not redraw:
        return
    progress  = float(i/imax)
    block     = int(round(bar_length * progress))
    endofline = '\r' if progress<1 else '\n'
    text = "{:16s} [{}] {:>5.1f}% of {}".format( what, "#"*block+"-"*(bar_length-block), progress*100, total_max)
    print(text, end=endofline)

    
def rmax(l):
    """
    Recursive max() for a given iterable of iterables
    Should be np.array of np.array or list of list, etc.
    args:
        l : Iterable of iterables
    return: 
        max value
    """
    maxi = float('-inf')
    for item in l:
        if isinstance(item, Iterable):
            t = rmax(item)
        else:
            t = item
        if t > maxi:
            maxi = t
    return maxi


def rmin(l):
    """
    Recursive min() for a given iterable of iterables
    Should be np.array of np.array or list of list, etc.
    args:
        l : Iterable of iterables
    return: 
        min value
    """
    mini = float('inf')
    for item in l:
        if isinstance(item, Iterable):
            t = rmin(item)
        else:
            t = item
        if t < mini:
            mini = t
    return mini

# def random_id(size=8, chars=string.ascii_letters + string.digits):
def random_id(size=8, chars='ABCDEFGHJKLMNPQRSTUVWXYZ0123456789abcdefghijkmnpqrstuvwxyz'):
    '''
    Return a random string id
    args:
        size  : id size (8)
        chars : Characters used
    return:
        id string
    '''
    return ''.join(random.choice(chars) for _ in range(size))



# -------------------------------------------------------------
# Fidle
# -------------------------------------------------------------

def get_datasets_dir(datasets_env_var='FIDLE_DATASETS_DIR'):
    '''
    Return dataset directory.
    If an env_var is defined, get, verify and return this defined path.
    If no env_var is defined, try to find a fidle datasets directory.
    If no valid datasets directory can be found, raise an Exception
    args:
        datasets_env_var : Environment var giving the location of the datasets
    return:
        datasets directory 
    '''

    # ---- Method 1 : env var
    #
    datasets_dir = os.getenv(datasets_env_var, None)
    if datasets_dir is not None:
        # Defined and exist, good
        if os.path.isdir(datasets_dir) : return datasets_dir
        # Defined, but cannot be found : erreur
        print('** OUPS !!\n----')
        print(f'** The directory specified by the {datasets_env_var} environment variable cannot be found...')
        print(f'** You should check this variable.\n')
        print('----')
        raise ValueError( f'datasets folder not found, please check {datasets_env_var} env var.' )

    # ---- Method 2 : We look around...
    #
    notebooks,datasets = looking_about()
    # No datasets dir found
    if len(datasets) == 0:
        print('** OUPS !!\n----')
        print(f'** No directory containing the datasets was found !')
        print('** The datasets are essential for this practical work.')
        print(f'** If you have installed datasets somewhere, you can specify the path via the {datasets_env_var} environment variable')
        print(f'** For more details, you can consult the installation procedure at : https://fidle.cnrs.fr/installation')
        print('----')
        raise ValueError( f'datasets folder not found, please check {datasets_env_var} env var.' )

    # ---- Get one, resolve tilde and return
    #
    datasets_dir = datasets[0]['path']
    datasets_dir = os.path.expanduser(datasets_dir)
    return  datasets_dir


def __looking_down_4about(path, depth):
    '''
    Search some fidle/about.yml files in a descending and recurring manner
    Note: Do not look further down when an instance is found
    args:
        path : path to explore
        p    : maximum depth
    '''
    # ---- Y a t-il un fichier about.yml ici ?
    if os.path.isfile(f'{path}/fidle/about.yml'): return [f'{path}/fidle/about.yml']
    # ---- Sinon, est-ce que l'on regarde dessous ?
    depth=depth-1
    # ---- Ben non, on est descendu assez bas
    if depth<0 : return []
    # ---- Ben oui, on regarde dessous
    abouts=[]
    dirs   = [d for d in glob(f'{path}/**', recursive=False) if os.path.isdir(d)]
    for d in dirs:
        sub_abouts=__looking_down_4about(d,depth)
        abouts.extend(sub_abouts)
    return abouts


def __looking_up_4about(path, up_depth, down_depth):
    '''
    Search fidle/about.yml files in an ascending and descending way, all around...
    args:
        path : starting point to explore
        up_depth   : maximum ascending depth
        down_depth : maximum descending path
    return:
        list of path, sorted and unique
    '''
    pp=Path(path)
    abouts=[]
    for i in range(up_depth):
        # print(pp)
        a=__looking_down_4about(pp, down_depth)
        abouts.extend(a)
        pp=pp.parent.absolute()
    # Supp. doublons + classement
    abouts = sorted( list(set(abouts)) )
    return abouts


def looking_about(path=None, up_depth=3, down_depth=3):
    '''
    Search fidle/about.yml files in an ascending and descending way 
    and extract informations from these files (Fidle internal use)
    args:
        path : starting point to explore, if None take current directory (None)
        up_depth   : maximum ascending depth (3)
        down_depth : maximum descending path (3)
    return:
        (notebooks, datasets) where notebooks, datasets are some array of descriptions
    '''
    # ---- Path None mean here
    #
    if path is None : path = os.getcwd()

    # ---- Clean path
    #
    path = os.path.expanduser(path)
    path = os.path.abspath(path)
    
    # ---- Get abbout files arround path
    #
    about_files = __looking_up_4about(path, up_depth,down_depth)
    
    # ---- Retrieve infos 
    #
    notebooks, datasets = [],[]
    for path in about_files:
        with open(path) as f:  about = yaml.full_load(f)
        path        = path.replace('/fidle/about.yml','')
        version     = about.get('version',     '??')
        content     = about.get('content',     '??')
        name        = about.get('name',        '??')
        description = about.get('description', '??')     

        entry = dict( path=path, version=version, content=content, name=name, description=description )

        if content == 'datasets':  datasets.append(entry)
        if content == 'notebooks': notebooks.append(entry)
    
    return notebooks,datasets
        



# ==================================================================
#                  _____ _                           
#                 / ____| |                          
#                | |    | |__  _ __ ___  _ __   ___  
#                | |    | '_ \| '__/ _ \| '_ \ / _ \ 
#                | |____| | | | | | (_) | | | | (_) |
#                 \_____|_| |_|_|  \___/|_| |_|\___/ 
#                                                    fidle.Chrono
# ==================================================================
# A simple class pour implement a stopwatch (and a bit more)
# Jean-Luc Parouty CNRS/MIAI/SIMaP 2022

import datetime
import fidle.config as config
import fidle.utils  as utils


class Chrono:

    def __init__(self):
        __version__   = config.VERSION
        self.reset_chrono()


    def start(self, id='default'):
        self.start_time[id] = datetime.datetime.now()


    def stop(self, id='default'):
        self.end_time[id]   = datetime.datetime.now()


    def get_delay(self, id='default', format='str'):
        t1 = self.start_time[id]
        t2 = self.end_time.get(id, datetime.datetime.now())
        dt = t2 - t1
        # ---- in seconds
        if format=='seconds': 
            return round(dt.total_seconds(),2)
        # ---- str, seconds without microseconds
        if format=='str':
            dt = dt - datetime.timedelta(microseconds=dt.microseconds)
            return str(dt)
        # ---- Human
        if format=='human':
            return utils.hdelay(dt)
        # ---- timedelta
        return dt


    @staticmethod
    def now(format='%d/%m/%y %H:%M:%S'):
        t = datetime.datetime.now()
        if format is None:  return t
        return t.strftime(format)
    

    @staticmethod
    def tag_now():
        return datetime.datetime.now().strftime("%Y-%m-%d_%Hh%Mm%Ss")


    def get_start(self, id='default', format='%d/%m/%y %H:%M:%S'):
        t1 = self.start_time[id]
        if format is None:  return t1
        return t1.strftime(format)


    def get_end(self, id='default', format='%d/%m/%y %H:%M:%S'):
        t2 = self.end_time[id]
        if format is None:  return t2
        return t2.strftime(format)

    def reset_chrono(self):
        self.start_time = {} 
        self.end_time   = {}

    def show(self,id='default'):
        print('Duration : ', self.get_delay(id,format='seconds'),'seconds')




# ==================================================================
#            ______ _     _ _        __  __           _ 
#           |  ____(_)   | | |      |  \/  |         | |
#           | |__   _  __| | | ___  | \  / | ___   __| |
#           |  __| | |/ _` | |/ _ \ | |\/| |/ _ \ / _` |
#           | |    | | (_| | |  __/ | |  | | (_) | (_| |
#           |_|    |_|\__,_|_|\___| |_|  |_|\___/ \__,_|
#                                                          
# ==================================================================
# A simple module to host usefull functions for Fidle practical work
# Jean-Luc Parouty CNRS/MIAI/SIMaP 2022

import os,sys,platform
import glob
import shutil
import pathlib

from random import randint
import matplotlib

from IPython.display import display,HTML

import fidle.config
import fidle.scrawler
import fidle.utils

from fidle.Chrono   import Chrono


__version__   = fidle.config.VERSION

__run_id        = None
__run_dir       = None
__datasets_dir  = None

utils           = fidle.utils
plot            = fidle.scrawler
config          = fidle.config
chrono          = Chrono()


# -------------------------------------------------------------
# init
# -------------------------------------------------------------
#
'''
Initialization with parameters run_id and run_dir.
These two parameters can be overridden via environment variables
FIDLE_OVERRIDE_run_id and FIDLE_OVERRIDE_run_dir.
datasets_dir is retrieved via the environment variable <datasets_env_var>.
params:
    run_id           : Run id for the notebook (random number if None)           (None)
    run_dir          : Run directory (./run/{run_id} if None)                    (None)
    datasets_env_var : Name of env. var. specifying the location of the datasets (FIDLE_DATASETS_DIR)
return:
    run_id, run_dir, datasets_dir
'''
def init(run_id=None, run_dir=None, datasets_env_var='FIDLE_DATASETS_DIR'):
    global __run_id
    global __run_dir
    global __datasets_dir
    
    # ---- run_id
    # 
    # If None, we choose
    __run_id    = str(randint(10000,99999)) if run_id  is None else run_id

    # Overrided ?
    __run_id    = __get_override_env('run_id', __run_id)

    # ---- run_dir 
    # 
    # If None, we set it
    __run_dir   = f'./run/{__run_id}'       if run_dir is None else run_dir

    # Override run_dir ?
    __run_dir   = __get_override_env('run_dir', __run_dir)
    
    # Create run_dir    
    utils.mkdir(__run_dir)

    # ---- Parameters from config.py
    #
    mplstyle  = config.FIDLE_MPLSTYLE
    cssfile   = config.FIDLE_CSSFILE
    
    # ---- Load matplotlib style and css
    #
    module_dir = pathlib.Path(__file__).parent
    matplotlib.style.use(f'{module_dir}/{mplstyle}')
    __load_cssfile(f'{module_dir}/{cssfile}')
    
    # ---- Get datasets location 
    #      From env var or by looking for
    #      Raise an exception if cannot be found.
    #
    __datasets_dir=utils.get_datasets_dir(datasets_env_var)
            
    # ---- Update Keras cache
    #
    updated = __update_keras_cache()

    # ---- Tensorflow log level
    #
    log_level = int(os.getenv('TF_CPP_MIN_LOG_LEVEL', 0 ))
    str_level = ['Info + Warning + Error','Warning + Error','Error only'][log_level]
    
    # ---- Today, now and hostname
    #
    chrono.start('__global__')
    h = platform.uname()
    
    # ---- Save figs
    #
    save_figs = os.getenv('FIDLE_SAVE_FIGS', str(config.SAVE_FIGS) )
    save_figs = (save_figs.lower() == 'true')
    figs_dir  = f'{__run_dir}/figs'

    plot.set_save_fig( save = save_figs, 
                  figs_dir  = figs_dir, 
                  figs_name = f'fig_{__run_id}_', 
                  figs_id   = 0 )

    # ---- Hello world
    #
    utils.display_md('<br>**FIDLE - Environment initialization**')
    print('Version              :', config.VERSION)
    print('Run id               :', __run_id)
    print('Run dir              :', __run_dir)
    print('Datasets dir         :', __datasets_dir)
    print('Start time           :', chrono.get_start('__global__'))
    print('Hostname             :', f'{h[1]} ({h[0]})')
    print('Tensorflow log level :', str_level,f' (={log_level})')
    print('Update keras cache   :', updated)
    print('Save figs            :', figs_dir, f'({save_figs})')
    
    # ---- Modules versions
    #
    for m in config.USED_MODULES:
        if m in sys.modules:
            print(f'{m:21s}:', sys.modules[m].__version__)

    # ---- Overrided ?
    #
    if run_id != __run_id:  
        print(f'\n** run_id has been overrided from {run_id} to {__run_id}')


    return __run_id, __run_dir, __datasets_dir

# ------------------------------------------------------------------
# Update keras cache
# ------------------------------------------------------------------
# Try to sync ~/.keras/cache with datasets/keras_cache
# because sometime, we cannot access to internet... (IDRIS..)
#
def __update_keras_cache():
    updated = False
    if os.path.isdir(f'{__datasets_dir}/keras_cache'):
        from_dir = f'{__datasets_dir}/keras_cache/*.*'
        to_dir   = os.path.expanduser('~/.keras/datasets')
        utils.mkdir(to_dir)
        for pathname in glob.glob(from_dir):
            filename=os.path.basename(pathname)
            destname=f'{to_dir}/{filename}'
            if not os.path.isfile(destname):
                shutil.copy(pathname, destname)
                updated=True
    return updated


# ------------------------------------------------------------------
# Override
# ------------------------------------------------------------------
#  
    
def override(*names, module_name='__main__', verbose=True, return_attributes=False):
    '''
    Try to override attributes given par name with environment variables.
    Environment variables name must be : FIDLE_OVERRIDE_<NAME>
    If no env variable is available for a given name, nothing is change.
    If type is str, substitution is done with 'notebook_id' and 'datasets_dir'
    Example : override('image_size','nb_epochs')
    params:
       names : list of attributes names as a str list
               if empty, all attributes can be override
    return :
       dict {name=new value}
    '''
    # ---- Where to override
    #
    module=sys.modules[module_name]
    
    # ---- No names : mean all
    #
    if len(names)==0:
        names=[]
        for name in dir(module):
            if name.startswith('_'): continue
            v=getattr(module,name)
            if type(v) not in [str, int, float, bool, tuple, list, dict]: continue
            names.append(name)
            
    # ---- Search for names
    #
    overrides={}
    for name in names:
        
        # ---- Variable doesn't exist
        #
        if not hasattr(module,name):
            print(f'** Warning : You try to override an inexistant variable ({name})')
            continue

        # ---- Get override environment variable name
        #
        env_value = __get_override_env(name, None)

        # ---- Environment variable doesn't exist
        #
        if env_value is None: continue

        # ---- Environment variable and variable exist both
        #
        value_old  = getattr(module,name)
        value_type = type(value_old)
        
        if value_type in [ str ] : 
            new_value = env_value.format(datasets_dir=__datasets_dir, run_id=__run_id)

        if value_type in [ int, float, bool, tuple, list, dict, type(None)]:
            new_value = eval(env_value)
    
        # ---- Override value
        #
        setattr(module,name,new_value)
        overrides[name]=new_value

    if verbose and len(overrides)>0:
        print('** Overrided parameters : **')
        for name,value in overrides.items():
            print(f'{name:20s} : {value}')
            
    if return_attributes:
        return overrides
       
  
def __get_override_env(name, default_value=None):
    env_name  = f'FIDLE_OVERRIDE_{name}'
    env_value = os.environ.get(env_name, default_value) 
    return env_value


# def __has_override_env(name):
#     env_name  = f'FIDLE_OVERRIDE_{name}'
#     return (env_name in os.environ)



def __load_cssfile(cssfile):
    if cssfile is None: return
    styles = open(cssfile, "r").read()
    display(HTML(styles))
    
     

    
     
def end():
    chrono.stop('__global__')
    end_time = chrono.get_end('__global__')
    duration = chrono.get_delay('__global__', format='human')
    site_url = "https://fidle.cnrs.fr"
    md = f'**End time :** {end_time}  \n'
    md+= f'**Duration :** {duration}  \n'
    md+= f'This notebook ends here :-)  \n'
    md+= f'[{site_url}]({site_url})'
    utils.display_md(md)
     
     



# ------------------------------------------------------------------
#   __  __             _      _____                       _   
#  |  \/  |           (_)    |  __ \                     | |  
#  | \  / | __ _  __ _ _  ___| |__) |___ _ __   ___  _ __| |_ 
#  | |\/| |/ _` |/ _` | |/ __|  _  // _ \ '_ \ / _ \| '__| __|
#  | |  | | (_| | (_| | | (__| | \ \  __/ |_) | (_) | |  | |_ 
#  |_|  |_|\__,_|\__, |_|\___|_|  \_\___| .__/ \___/|_|   \__|
#                 __/ |                 | |                   
#                |___/                  |_|        fidle.MagicReport
# ------------------------------------------------------------------
# A simple class to manage execution reports
# Jean-Luc Parouty CNRS/MIAI/SIMaP 2022

import os
import pathlib
import json
from string import Template

import fidle.config as config


class MagicReport :

    __version__  = config.VERSION
    

    def __init__(self, campain=None, notebooks=None, task_file=None):
        self.campain   = campain
        self.notebooks = notebooks
        self.task_file = task_file
        self.template  = campain.get('report_template','default')
        


    # --------------------------------------------------------------
    # -- Report stuffs
    # --------------------------------------------------------------
    #
    def build_report(self):

        # ---- Retrieve infos
        #
        campain   = self.campain
        notebooks = self.notebooks

        with open(self.task_file) as fp:
                task_list = json.load(fp)

        # ---- Metadata part
        #
        html_metadata=''
        for t in ['version', 'description', 'date', 'host', 'profile']:
            html_metadata +=  f'<b>{t.title()}</b> : '  +  campain[t] +'<br>\n'

        # ---- Retrieve reports
        #
        report_dir    = campain['campain_html_dir']
        feedback_dir  = campain['campain_feedback_dir']
        missing       = []

        # ---- Add table header
        html_report = '<table>'
        html_report += '<tr><th>Id</th><th>Status</th><th>Notebook</th><th>Start</th><th>Duration</th><th>Running</th><th>By</th></tr>\n'

        # ---- Add lines
        #
        for task_id in notebooks.keys():
            feedback_file = f'{feedback_dir}/{task_id}.json'
            
            # ---- Get report, if exist
            #
            if os.path.isfile(feedback_file):
                # Read it 
                with open(feedback_file) as fp:
                    feedback = json.load(fp)
            else:
                # Empty report
                feedback={}
                missing.append(task_id)

            # ---- Get task if exist
            #
            task={}
            for t in task_list:
                if t['id']==task_id: task=t         

            # ---- Get infos
            #
            s_notebook = feedback.get('notebook', {})
            s_feedback = feedback.get('feedback', {})

            path        = s_notebook.get('notebook', '-')
            status      = s_feedback.get('status',   '-')
            start       = s_feedback.get('start',    '-')
            duration    = s_feedback.get('duration', '-')
            done_html   = s_feedback.get('output_html_file', '#')
            task_status = task.get('status','-')
            task_runner = task.get('runner','-')
                   
            if task_status not in ['running','done','ready'] : 
                task_status = '-'
                task_runner = '-'
            
            cols = []
            cols.append( task_id  )
            cols.append( status   )
            cols.append( f'<a href="{done_html}" target="_blank">{path}</a>' )
            cols.append( start    )
            cols.append( duration )
            cols.append( task_status )
            cols.append( task_runner )

            html_report+='\n<tr>'
            for c in cols:
                html_report+=f'<td>{c}</td>'
            html_report+='</tr>\n'

        html_report+='</table>'

        # ---- Finalize report
        #
        body_html = self.__finalize_report(html_metadata, html_report)

        # ---- Save it
        #
        report_file = f'{report_dir}/index.html'

        with open(report_file, "wt") as fp:
            fp.write(body_html)
        # print(f'\nSaved HTML full report : {report_file}')
            

    

    def __finalize_report(self,html_metadata, html_report):

        # ---- Load template
        # 
        if self.template in ['default', 'fidle']:
            # Predefined template
            module_dir = pathlib.Path(__file__).parent
            template_path = f'{module_dir}/templates/MagicReport_{self.template}.html'
        else:
            # Given template
            template_path = self.template

        template = Template( open(template_path).read() )

        # ---- Template substitution
        #
        if self.template=='fidle':
            logo_header = open(f'{module_dir}/img/00-Fidle-header-01.svg'   ).read()
            logo_ender  = open(f'{module_dir}/img/00-Fidle-logo-01-80px.svg').read()
            html= template.substitute( logo_header   = logo_header, 
                                       logo_ender    = logo_ender,
                                       html_report   = html_report, 
                                       html_metadata = html_metadata,
                                       fidle_version = config.VERSION)
        else:
            html= template.substitute( html_report   = html_report, 
                                       html_metadata = html_metadata,
                                       fidle_version = config.VERSION)

        return html


# ------------------------------------------------------------------
#       _______ _                 _____ _               _   
#      |__   __(_)               / ____| |             | |  
#         | |   _ _ __ ___   ___| (___ | |__   ___  ___| |_ 
#         | |  | | '_ ` _ \ / _ \\___ \| '_ \ / _ \/ _ \ __|
#         | |  | | | | | | |  __/____) | | | |  __/  __/ |_ 
#         |_|  |_|_| |_| |_|\___|_____/|_| |_|\___|\___|\__|
#
#                                            fidle.TimeSheetManager         
# ------------------------------------------------------------------
#     
# A simple class to manage time sheets.
# Jean-Luc Parouty CNRS/MIAI/SIMaP 2022



import os,platform,sys
import json
import datetime, time
import dateutil
import pathlib
import base64
import hashlib
from random  import uniform
from string import Template
import yaml

from IPython.display import display
import pandas as pd

from email.header         import decode_header
from email.mime.multipart import MIMEMultipart
from email.mime.text      import MIMEText
import email
import imaplib
import smtplib
import fidle.utils


import fidle.config
from fidle.Chrono import Chrono


class TimeSheetManager:

    __version__  = fidle.config.VERSION


    def __read_profile(self, profile_path):
        '''
        Load profile
        args:
            profile_name : path to the yaml profile file
        return:
            profile
        '''
        profile_path =  os.path.expanduser(profile_path)
        with open(profile_path,'r') as fp:
            profile=yaml.load(fp, Loader=yaml.FullLoader)
        return profile


    # --------------------------------------------------------------
    # -- Init TimeSheetManager
    # --------------------------------------------------------------
    #
    def __init__(self, profile_path, debug=0):
        '''
        Init TimeSheetManager
        args:
            profile_path : Path to the profile file
            debug : debug level 0:none 1:debug traces (0)
        return:
            nothing
        '''

        self.__debug_level = debug

        print(f'\n=== Fidle/TimeSheetManager - Management of presence sheets')
        print(f'=== Version : {self.__version__}\n')

        # ---- Load profile
        #
        profile = self.__read_profile(profile_path)

        # ---- Retrieve infos (sequences key must be titled)
        #
        self.config     = profile['config']
        self.sequences  = { k.title():v for (k,v) in profile['sequences'].items() }
        
        print('TimeSheetManager init')
        print(f'    Profile         : {profile_path}')
        print(f"    Mailto          : {self.config['mail_to']}")
        print(f'    Sequences       : {len(self.sequences)}\n')



    def __open_M(self):
        '''
        Open IMAP session
        '''
        imap_server  = self.config['imap_server']
        imap_login   = self.config['imap_login']
        imap_passwd  = self.config['imap_password']

        M = imaplib.IMAP4_SSL(imap_server)
        M.login(imap_login, imap_passwd)

        self.__M = M


    def __close_M(self):
        '''
        Close IMAP session
        '''
        self.__M.close()
        self.__M.logout()


    def __debug(self,*what):
        '''
        Print debug infos, if debug level is >0..
        args:
            *what: stuffs to print
        '''
        if self.__debug_level <1: return
        for w in what:
            print(w)


    
    def __open_S(self):
        '''
        Open SMTP session
        '''

        smtp_server    = self.config['smtp_server']
        smtp_port      = self.config['smtp_port']
        smtp_login     = self.config['smtp_login']
        smtp_passwd    = self.config['smtp_password']

        S = smtplib.SMTP(smtp_server,port=smtp_port)
        S.ehlo()
        S.starttls() 
        S.login(smtp_login, smtp_passwd)

        self.__S = S


    def __close_S(self):
        '''
        Close SMTP session
        '''
        self.__S.quit()



    def __check_imap_status(self,status,data):
        '''
        Check imaplib return status. Raise an exception if not ok.
        '''
        if status != 'OK' : raise RuntimeError(data[0])



    def __visa_encode(self, email, sequence_id, magic_token, digest_size=20):
        '''
        Encode a visa from email and sequence_id
        args:
            email : email of participant
            sequence_id : followed sequence_id
        return:
            base32 visa
        '''
        secret_text = f'{email.lower()} {magic_token} {sequence_id.lower()}'

        h = hashlib.blake2b(digest_size=digest_size)
        h.update(secret_text.encode('UTF8'))
        visa_b = h.digest()
        visa = base64.b32encode(visa_b).decode('ascii').upper()
        return visa



    def __get_confirmation(self, presence):
        
        module_dir       = pathlib.Path(__file__).parent
        template         = self.config['template']
        mail_from        = self.config['mail_from']
        recovery_url     = self.config['recovery_url']

        mail_to          = presence.get('from')
        attestation_visa = presence.get('attestation_visa')
        sequence_id      = presence.get('sequence_id')

        encoded = base64.b64encode(open(f'{module_dir}/img/00-Fidle-logo-01_m.png', "rb").read()).decode('utf-8')
        logo_ender = '<img src="data:image/png;base64,'+encoded+'"/>'

        recovery_url = recovery_url.format( mail_id          = mail_to, 
                                            sequence_id      = sequence_id, 
                                            attestation_visa = attestation_visa)

        # ---- Load template
        # 
        if template in ['default', 'fidle']:
            # Predefined template
            template_path = f'{module_dir}/templates/Confirmation_{template}'
        else:
            # Given template
            template_path = self.template

        template_txt  = Template( open(template_path+'.txt' ).read() )
        template_html = Template( open(template_path+'.html').read() )

        # ---- Template substitution
        #
        if template=='fidle':
            subject = 'Fidle / Confirmation' 
            html    = template_html.substitute( logo_ender       = logo_ender,
                                                recovery_url     = recovery_url,
                                                mail_id          = mail_to, 
                                                attestation_visa = attestation_visa,
                                                fidle_version    = self.__version__)
            plain   = template_txt.substitute(  mail_id          = mail_to, 
                                                attestation_visa = attestation_visa,
                                                fidle_version    = self.__version__)
        else:
            subject = 'Confirmation' 
            html    = template_html.substitute( mail_id          = mail_to, 
                                                attestation_visa = attestation_visa,
                                                fidle_version    = self.__version__)
            plain   = template_txt.substitute(  mail_id          = mail_to, 
                                                attestation_visa = attestation_visa,
                                                fidle_version    = self.__version__)
        
        # ---- Create mail
        #
        # Create message container -  MIME is multipart/alternative.
        #
        msg = MIMEMultipart('alternative')
        msg['Subject'] = subject
        msg['From']    = mail_from
        msg['To']      = mail_to

        # Two parts : text/plain and text/html.
        #
        part1 = MIMEText(plain, 'plain')
        part2 = MIMEText(html,  'html' )

        # Attach parts into message container.
        # According to RFC 2046, the last part of a multipart message is best and preferred

        msg.attach(part1)
        msg.attach(part2)

        return msg




    def get_presences( self, 
                       sequence_id = None,
                       magickey    = None, 
                       at          = None, 
                       duration    = 15):
        '''
        Retreive presences mail for a given sequence
        args:
            presence_day : Day of the event
            magickey : Magic key to search
            show_all : See all mails or only the ones that are validated
        return:
            dict of mails
        '''
        self.__debug('== __get_presences() ======')

        print('=== Get presences from mails ===\n')

        # ---- Prepare parameters
        #
        imap_inbox   = self.config['imap_inbox']
        mail_to      = self.config['mail_to']
        magic_token  = self.config['magic_token']
        digest_size  = self.config['digest_size']

        tz_local     = dateutil.tz.gettz(self.config['tz_local'])

        # Check
        #
        if sequence_id is None and magickey is None:
            raise('Sequence id or magickey must be defined...') 
        
        # Sequence
        #
        if sequence_id is None:
            sequence_id    = '-'
            sequence_title = '-'
            sequence_date  = '-'
        else:
            sequence_id    = sequence_id.title()
            sequence_title = self.sequences[sequence_id]['title']
            sequence_date  = self.sequences[sequence_id]['date']

        # Magic key
        #
        if magickey is None: 
            magickey = self.sequences[sequence_id.title()]['magickey']

        # t1 from at
        #
        if at is None:
            t1 = datetime.datetime.now()
            t1 = t1.astimezone(tz_local)
        else:
            t1 = datetime.datetime.strptime(at, '%d-%m-%Y %H:%M:%S')
            t1 = t1.astimezone(tz_local)

        # t2 from t1 + duration
        #
        t2 = t1 + datetime.timedelta(minutes=duration)

        # after and before
        #
        after  = min(t1,t2)
        before = max(t1,t2)

        # ---- Verbosity
        #
        print('Parameters are :')
        print(f'    Sequence id     : {sequence_id}')
        print(f'    Sequence title  : {sequence_title}')
        print(f'    Sequence date   : {sequence_date}')
        print(f'    Magickey        : {magickey}')
        print(f"    valid after     : {after.strftime('%d-%m-%Y %H:%M:%S')}")
        print(f"    valid before    : {before.strftime('%d-%m-%Y %H:%M:%S')}\n")

        # ---- Open IMAP channel
        #
        self.__open_M()
        M= self.__M
        presences = []

        # ---- Build filter
        #
        # filter = f'OR TEXT "{magickey}" SUBJECT "{magickey}" ON "{presence_day}" TO "{mail_to}"'
        # filter = f'OR TEXT "{magickey}" SUBJECT "{magickey}" TO "{mail_to}"'
        filter = f'TO "{mail_to}"'

        # ---- Go to the right box, and search messages
        #
        M.select(imap_inbox)
        # status, data = M.search(None, filter)

        status, uids = M.uid('SEARCH', filter)

        self.__check_imap_status(status,uids)
        self.__debug(f'folder={imap_inbox}',f'filter={filter}',f'status={status}')

        # ---- Split uids and get messages
        #
        mail_uids = []
        for block in uids:
            mail_uids += block.split()
        self.__debug(f'Nb mails={len(mail_uids)}')

        nb_ok = 0
        for uid in mail_uids:

            uid=uid.decode('UTF8')
            # ---- Get Flags
            # status,  flags = M.fetch(id, '(FLAGS)')
            status,  flags = M.uid('FETCH', uid, '(FLAGS)')
            self.__check_imap_status(status,flags)

            flags = flags[0].decode('UTF8')
            self.__debug(f'flags={flags}')

            flag_answered = ( '\Answered' in flags)
            flag_seen     = ( '\Seen'     in flags)

            # ---- Get message
            #
            # status,  data  = M.fetch(id, '(RFC822)')
            status,  data  = M.uid('FETCH',uid, '(RFC822)')
            self.__check_imap_status(status,data)

            # Note : data is a multiparts data setructure
            # Dans notre cas, on va considérer que nous n'avons qu'une seule partie.
            # Pour être rigoureux, il faudrait itérer sur les morceaux...
            #
            # for response_part in data:
            #     if not isinstance(response_part, tuple): continue
            #     message     = email.message_from_bytes(response_part[1])

            message     = email.message_from_bytes( data[0][1] )

            msg_to      = email.utils.parseaddr( message.get('to')   )[1]
            msg_from    = email.utils.parseaddr( message.get('from') )[1]
            msg_subject = message.get('subject')
            msg_id      = message.get('Message-ID')

            # ---- Extract content from brut message
            #
            if message.is_multipart():
                msg_content = ''

                for part in message.get_payload():

                    if part.get_content_type() == 'text/plain':
                        msg_content += part.get_payload()
            else:
                msg_content = message.get_payload()

            # ---- Is it a presence message ? (have magickey)
            #
            msg_magic_content =  ( magickey in msg_content )
            msg_magic_subject =  ( magickey in msg_subject )

            # ---- Send date
            #
            msg_send_str = message.get('date')
            msg_send_dt  = email.utils.parsedate_to_datetime(msg_send_str)
            msg_send_dt  = msg_send_dt.astimezone(tz_local)
            msg_send_ok  = (after <= msg_send_dt <= before)
            # print('Send date : ', msg_send_dt.strftime('%A %d %b %Y %H:%M:%S'), msg_send_ok)


            # ---- Received date
            #
            msg_received_str = message.get('received').split(';')[-1]
            msg_received_dt  = email.utils.parsedate_to_datetime(msg_received_str)
            # print('Rec date  : ', msg_received_dt.strftime('%A %d %b %Y %H:%M:%S'))

            # ---- Transit
            #
            dt = msg_received_dt - msg_send_dt
            msg_transit = dt.total_seconds()
            # print(f'Transit   : {msg_transit:.2f}')

            # ---- Age
            #
            now = datetime.datetime.now().astimezone(tz_local)
            dt =  now - msg_send_dt
            msg_age = dt.total_seconds()
            # print(f'Age/send  : {msg_age:.2f} sec.')

            # ---- Check validity
            #      Need valid send date and magic content
            #
            attestation_ok = (msg_magic_content or msg_magic_subject) and msg_send_ok
            if attestation_ok: nb_ok+=1

            # ---- Set visa
            #
            attestation_visa = self.__visa_encode(msg_from, sequence_id, magic_token, digest_size) if attestation_ok else '-'

            # ---- Create presence
            #
            presence = { 'imap_uid'         : uid,
                         'msg_id'           : msg_id,
                         'from'             : msg_from,
                         'to'               : msg_to,
                         'magic_subject'    : msg_magic_subject,
                         'magic_content'    : msg_magic_content,
                         'send_dt'          : msg_send_dt,
                         'send_str'         : msg_send_str,
                         'send_ok'          : msg_send_ok,
                         'received_dt'      : msg_received_dt,
                         'received_str'     : msg_received_str,
                         'transit'          : msg_transit,
                         'age'              : msg_age,
                         'flag_answered'    : flag_answered,
                         'flag_seen'        : flag_seen,
                         'attestation_ok'   : attestation_ok,
                         'attestation_visa' : attestation_visa,
                         'sequence_id'      : sequence_id
            }

            presences.append(presence)

        self.__close_M()

        # ---- Verbosity
        #
        print('Presences :')
        print(f'    Found in folder : {len(presences)}')
        print(f'    Attestations ok : {nb_ok}\n')

        return presences



    def send_confirmations(self, presences, save_copy=False, flag_answered=False):
        '''
        Sending confirmations for the given presence list.
        Confirmations are sent only if attestation_ok is true.
        args:
            presences     : presences list
            save_copy     : Save a copy of each confirmation email sent (False)
            flag_answered : Set the answered flag of the emails to which a confirmation is sent (False)
        returns:
            nothing
        '''

        imap_inbox     = self.config['imap_inbox']
        imap_sentbox   = self.config['imap_sentbox']

        print('=== Send confirmations ===\n')
        print('[x]: Invalid message  [o]: Confirmation sent  [.]:Previously answered\n')

        # ---- Open SMTP session
        #
        self.__open_S()
        S=self.__S

        # ---- Open IMAP session
        #
        self.__open_M()
        M= self.__M

        # ---- Iterate on presences
        #
        nb_sent=0
        for p in presences:
        
            # ---- Attestation not ok
            #
            if not p.get('attestation_ok',False):
                print('[x]', end='')
                continue
            
            # ---- Attestation ok, but already answered
            #
            if p.get('flag_answered'):
                print('[.]', end='')
                continue

            # ---- Attestation ok, answer to do
            #
            # Prepare message
            message = self.__get_confirmation(p)

            # Send it
            S.send_message(message)
            nb_sent+=1

            # Save a copy ?
            if save_copy:
                M.append( imap_sentbox, 
                          '\Seen', 
                          imaplib.Time2Internaldate(time.time()), 
                          str(message).encode('utf-8') )

            # Flag answered ?
            if flag_answered : 
                M.select( imap_inbox    )
                uid=p['imap_uid']
                M.uid( 'STORE', uid, '+FLAGS', '(\\Answered)')
                p['flag_answered']=True
                
            # Trace
            print('[o]', end='')
                
        print('\n')
        self.__close_S()

        print(f'Number of confirmations sent : {nb_sent}\n')

        return



    def show_uids(self):

        imap_inbox     = self.config['imap_inbox']

        self.__open_M()
        M= self.__M

        M.select( imap_inbox)
        uids = M.uid('SEARCH', 'ALL')
        print(uids)



    def unflag(self, presences):
        '''
        Remove anwered and seen flags from given presence emails.
        Only emails with a valid send date are updated.
        args:
            presences : list of presences
        return
            number of changed presence emails
        '''

        print('=== Remove flags ===\n')

        imap_inbox     = self.config['imap_inbox']

        self.__open_M()
        M= self.__M

        M.select( imap_inbox )

        nb_changed=0
        for p in presences:
            if p['send_ok'] and ( p['flag_answered'] or p['flag_seen'] ):
                uid=p['imap_uid']
                M.uid( 'STORE', uid, '-FLAGS', '\\Answered \\seen')
                p['flag_seen']     = False
                p['flag_answered'] = False
                nb_changed+=1

        print(f'Number of unflag emails : {nb_changed}\n')



    def save_presences(self, presences, save_as=None, default='Presences'):
        '''
        Save given presences as a nice csv file
        if save_as is None : do nothing
        is save_as is 'default' : use default base name
        args:
            presences : presences data
            save_as : filename to use, None or 'default'
            default : default base name to use if save_as is 'default'
        return:
            nothing
        '''

        # No need to save
        if save_as is None: return

        # Define filename
        if save_as=='default':
            filename=f'{default}-{Chrono.tag_now()}.csv'
        else:
            filename=save_as

        # Save it
        print('=== Save presences sheet ===\n')
        df = pd.DataFrame.from_dict(presences)
        df.to_csv(filename,sep=';')
        print(f'    Saved as     : {filename}')


        # df = df[ ['from','send_str','send_dt','transit','attestation_ok','sequence_id'] ]
        # with pd.option_context('display.max_rows', None, 'display.max_columns', None, 'display.width', 2000):
        #     display(df)
